@@ -1,5 +1,7 @@
-from math import floor
 from functools import reduce
+from heapq import (heappop,
+                   heappush)
+from math import floor
 from typing import (Iterator,
                     List,
                     Optional,
@@ -7,7 +9,6 @@ from typing import (Iterator,
                     Tuple,
                     Type)
 
-from prioq.base import PriorityQueue
 from reprit.base import generate_repr
 
 from .core import (hilbert as _hilbert,
@@ -185,38 +186,9 @@ class Tree:
         """
         return self._max_children
 
-    def find_interval_indices(self, interval: Interval) -> List[int]:
+    def find_subsets(self, interval: Interval) -> List[Interval]:
         """
-        Searches for indices of intervals
-        that lie inside the given closed interval.
-
-        Time complexity:
-            ``O(max_children * log size + hits_count)``
-        Memory complexity:
-            ``O(max_children * log size + hits_count)``
-
-        where ``size = len(self.intervals)``,
-        ``max_children = self.max_children``,
-        ``hits_count`` --- number of found indices.
-
-        :param interval: input interval.
-        :returns: indices of intervals that lie inside the input interval.
-
-        >>> intervals = [((-index, index), (0, index))
-        ...              for index in range(1, 11)]
-        >>> tree = Tree(intervals)
-        >>> tree.find_interval_indices(((-1, 1), (0, 1))) == [0]
-        True
-        >>> tree.find_interval_indices(((-2, 2), (0, 2))) == [0, 1]
-        True
-        >>> tree.find_interval_indices(((-3, 3), (0, 3))) == [0, 1, 2]
-        True
-        """
-        return [index for index, _ in self._find_interval_items(interval)]
-
-    def find_interval_intervals(self, interval: Interval) -> List[Interval]:
-        """
-        Searches for intervals that lie inside the given closed interval.
+        Searches for intervals that lie inside the given interval.
 
         Time complexity:
             ``O(max_children * log size + hits_count)``
@@ -233,23 +205,49 @@ class Tree:
         >>> intervals = [((-index, index), (0, index))
         ...              for index in range(1, 11)]
         >>> tree = Tree(intervals)
-        >>> (tree.find_interval_intervals(((-1, 1), (0, 1)))
-        ...  == [((-1, 1), (0, 1))])
+        >>> tree.find_subsets(((-1, 1), (0, 1))) == [((-1, 1), (0, 1))]
         True
-        >>> (tree.find_interval_intervals(((-2, 2), (0, 2)))
+        >>> (tree.find_subsets(((-2, 2), (0, 2)))
         ...  == [((-1, 1), (0, 1)), ((-2, 2), (0, 2))])
         True
-        >>> (tree.find_interval_intervals(((-3, 3), (0, 3)))
+        >>> (tree.find_subsets(((-3, 3), (0, 3)))
         ...  == [((-1, 1), (0, 1)), ((-2, 2), (0, 2)), ((-3, 3), (0, 3))])
         True
         """
         return [interval
-                for _, interval in self._find_interval_items(interval)]
+                for _, interval in self._find_subsets_items(interval)]
 
-    def find_interval_items(self, interval: Interval) -> List[Item]:
+    def find_subsets_indices(self, interval: Interval) -> List[int]:
         """
-        Searches for indices with intervals
-        that lie inside the given closed interval.
+        Searches for indices of intervals that lie inside the given interval.
+
+        Time complexity:
+            ``O(max_children * log size + hits_count)``
+        Memory complexity:
+            ``O(max_children * log size + hits_count)``
+
+        where ``size = len(self.intervals)``,
+        ``max_children = self.max_children``,
+        ``hits_count`` --- number of found indices.
+
+        :param interval: input interval.
+        :returns: indices of intervals that lie inside the input interval.
+
+        >>> intervals = [((-index, index), (0, index))
+        ...              for index in range(1, 11)]
+        >>> tree = Tree(intervals)
+        >>> tree.find_subsets_indices(((-1, 1), (0, 1))) == [0]
+        True
+        >>> tree.find_subsets_indices(((-2, 2), (0, 2))) == [0, 1]
+        True
+        >>> tree.find_subsets_indices(((-3, 3), (0, 3))) == [0, 1, 2]
+        True
+        """
+        return [index for index, _ in self._find_subsets_items(interval)]
+
+    def find_subsets_items(self, interval: Interval) -> List[Item]:
+        """
+        Searches for indices with intervals that lie inside the given interval.
 
         Time complexity:
             ``O(max_children * log size + hits_count)``
@@ -266,23 +264,119 @@ class Tree:
         >>> intervals = [((-index, index), (0, index))
         ...              for index in range(1, 11)]
         >>> tree = Tree(intervals)
-        >>> (tree.find_interval_items(((-1, 1), (0, 1)))
+        >>> (tree.find_subsets_items(((-1, 1), (0, 1)))
         ...  == [(0, ((-1, 1), (0, 1)))])
         True
-        >>> (tree.find_interval_items(((-2, 2), (0, 2)))
+        >>> (tree.find_subsets_items(((-2, 2), (0, 2)))
         ...  == [(0, ((-1, 1), (0, 1))), (1, ((-2, 2), (0, 2)))])
         True
-        >>> (tree.find_interval_items(((-3, 3), (0, 3)))
+        >>> (tree.find_subsets_items(((-3, 3), (0, 3)))
         ...  == [(0, ((-1, 1), (0, 1))), (1, ((-2, 2), (0, 2))),
         ...      (2, ((-3, 3), (0, 3)))])
         True
         """
-        return list(self._find_interval_items(interval))
+        return list(self._find_subsets_items(interval))
 
-    def _find_interval_items(self, interval: Interval) -> Iterator[Item]:
+    def find_supersets(self, interval: Interval) -> List[Interval]:
+        """
+        Searches for intervals that contain the given interval.
+
+        Time complexity:
+            ``O(max_children * log size + hits_count)``
+        Memory complexity:
+            ``O(max_children * log size + hits_count)``
+
+        where ``size = len(self.intervals)``,
+        ``max_children = self.max_children``,
+        ``hits_count`` --- number of found intervals.
+
+        :param interval: input interval.
+        :returns: intervals that contain the input interval.
+
+        >>> intervals = [((-index, index), (0, index))
+        ...              for index in range(1, 11)]
+        >>> tree = Tree(intervals)
+        >>> tree.find_supersets(((-10, 10), (0, 10))) == [((-10, 10), (0, 10))]
+        True
+        >>> (tree.find_supersets(((-9, 9), (0, 9)))
+        ...  == [((-9, 9), (0, 9)), ((-10, 10), (0, 10))])
+        True
+        >>> (tree.find_supersets(((-8, 8), (0, 8)))
+        ...  == [((-8, 8), (0, 8)), ((-9, 9), (0, 9)), ((-10, 10), (0, 10))])
+        True
+        """
+        return [interval
+                for _, interval in self._find_supersets_items(interval)]
+
+    def find_supersets_indices(self, interval: Interval) -> List[int]:
+        """
+        Searches for indices of intervals that contain the given interval.
+
+        Time complexity:
+            ``O(max_children * log size + hits_count)``
+        Memory complexity:
+            ``O(max_children * log size + hits_count)``
+
+        where ``size = len(self.intervals)``,
+        ``max_children = self.max_children``,
+        ``hits_count`` --- number of found indices.
+
+        :param interval: input interval.
+        :returns: indices of intervals that contain the input interval.
+
+        >>> intervals = [((-index, index), (0, index))
+        ...              for index in range(1, 11)]
+        >>> tree = Tree(intervals)
+        >>> tree.find_supersets_indices(((-10, 10), (0, 10))) == [9]
+        True
+        >>> tree.find_supersets_indices(((-9, 9), (0, 9))) == [8, 9]
+        True
+        >>> tree.find_supersets_indices(((-8, 8), (0, 8))) == [7, 8, 9]
+        True
+        """
+        return [index for index, _ in self._find_supersets_items(interval)]
+
+    def find_supersets_items(self, interval: Interval) -> List[Item]:
+        """
+        Searches for indices with intervals
+        that contain the given interval.
+
+        Time complexity:
+            ``O(max_children * log size + hits_count)``
+        Memory complexity:
+            ``O(max_children * log size + hits_count)``
+
+        where ``size = len(self.intervals)``,
+        ``max_children = self.max_children``,
+        ``hits_count`` --- number of found indices with intervals.
+
+        :param interval: input interval.
+        :returns: indices with intervals that contain the input interval.
+
+        >>> intervals = [((-index, index), (0, index))
+        ...              for index in range(1, 11)]
+        >>> tree = Tree(intervals)
+        >>> (tree.find_supersets_items(((-10, 10), (0, 10)))
+        ...  == [(9, ((-10, 10), (0, 10)))])
+        True
+        >>> (tree.find_supersets_items(((-9, 9), (0, 9)))
+        ...  == [(8, ((-9, 9), (0, 9))), (9, ((-10, 10), (0, 10)))])
+        True
+        >>> (tree.find_supersets_items(((-8, 8), (0, 8)))
+        ...  == [(7, ((-8, 8), (0, 8))), (8, ((-9, 9), (0, 9))),
+        ...      (9, ((-10, 10), (0, 10)))])
+        True
+        """
+        return list(self._find_supersets_items(interval))
+
+    def _find_subsets_items(self, interval: Interval) -> Iterator[Item]:
         yield from (enumerate(self._intervals)
                     if _interval.is_subset_of(self._root.interval, interval)
-                    else _find_node_interval_items(self._root, interval))
+                    else _find_node_interval_subsets_items(self._root,
+                                                           interval))
+
+    def _find_supersets_items(self, interval: Interval) -> Iterator[Item]:
+        yield from _find_node_interval_supersets_items(self._root, interval)
 
     def n_nearest_indices(self, n: int, point: Point) -> Sequence[int]:
         """
@@ -384,15 +478,17 @@ class Tree:
                     else enumerate(self._intervals))
 
     def _n_nearest_items(self, n: int, point: Point) -> Iterator[Item]:
-        queue = PriorityQueue((0, 0, self._root))
+        queue = [(0, 0, self._root)]
         while n and queue:
-            node = queue.pop()[2]
+            _, _, node = heappop(queue)
             for child in node.children:
-                queue.push((child.distance_to_point(point),
-                            -child.index - 1 if child.is_leaf else child.index,
-                            child))
-            while n and queue and queue.peek()[1] < 0:
-                yield queue.pop()[2].item
+                heappush(queue,
+                         (child.distance_to_point(point),
+                          -child.index - 1 if child.is_leaf else child.index,
+                          child))
+            while n and queue and queue[0][1] < 0:
+                _, _, node = heappop(queue)
+                yield node.item
                 n -= 1
 
     def nearest_index(self, point: Point) -> int:
@@ -475,15 +571,17 @@ class Tree:
         >>> tree.nearest_item((10, 10)) == (9, ((-10, 10), (0, 10)))
         True
         """
-        queue = PriorityQueue((0, 0, self._root))
+        queue = [(0, 0, self._root)]
         while queue:
-            node = queue.pop()[2]
+            _, _, node = heappop(queue)
             for child in node.children:
-                queue.push((child.distance_to_point(point),
-                            -child.index - 1 if child.is_leaf else child.index,
-                            child))
-            if queue and queue.peek()[1] < 0:
-                return queue.pop()[2].item
+                heappush(queue,
+                         (child.distance_to_point(point),
+                          -child.index - 1 if child.is_leaf else child.index,
+                          child))
+            if queue and queue[0][1] < 0:
+                _, _, node = heappop(queue)
+                return node.item
 
 
 def _create_root(intervals: Sequence[Interval],
@@ -497,19 +595,21 @@ def _create_root(intervals: Sequence[Interval],
         # only one node, skip sorting and just fill the root box
         return node_cls(len(nodes), interval, nodes)
     else:
-        (x_min, x_max), (y_min, y_max) = interval
+        (tree_min_x, tree_max_x), (tree_min_y, tree_max_y) = interval
 
         def node_key(node: Node,
-                     double_tree_delta_x: Coordinate = 2 * (x_max - x_min),
-                     double_tree_delta_y: Coordinate = 2 * (y_max - y_min),
-                     double_tree_x_min: Coordinate = 2 * x_min,
-                     double_tree_y_min: Coordinate = 2 * y_min) -> int:
-            (x_min, x_max), (y_min, y_max) = node.interval
+                     double_tree_delta_x: Coordinate
+                     = 2 * (tree_max_x - tree_min_x),
+                     double_tree_delta_y: Coordinate
+                     = 2 * (tree_max_y - tree_min_y),
+                     double_tree_min_x: Coordinate = 2 * tree_min_x,
+                     double_tree_min_y: Coordinate = 2 * tree_min_y) -> int:
+            (min_x, max_x), (min_y, max_y) = node.interval
             return _hilbert.index(floor(_hilbert.MAX_COORDINATE
-                                        * (x_min + x_max - double_tree_x_min)
+                                        * (min_x + max_x - double_tree_min_x)
                                         / double_tree_delta_x),
                                   floor(_hilbert.MAX_COORDINATE
-                                        * (y_min + y_max - double_tree_y_min)
+                                        * (min_y + max_y - double_tree_min_y)
                                         / double_tree_delta_y))
 
         nodes = sorted(nodes,
@@ -546,11 +646,21 @@ def _node_to_leaves(node: Node) -> Iterator[Node]:
             yield from _node_to_leaves(child)
 
 
-def _find_node_interval_items(node: Node,
-                              interval: Interval) -> Iterator[Item]:
+def _find_node_interval_subsets_items(node: Node,
+                                      interval: Interval) -> Iterator[Item]:
     if _interval.is_subset_of(node.interval, interval):
-        for node_leaf in _node_to_leaves(node):
-            yield node_leaf.item
-    elif not node.is_leaf and _interval.overlap(interval, node.interval):
+        for leaf in _node_to_leaves(node):
+            yield leaf.item
+    elif not node.is_leaf and _interval.overlaps(interval, node.interval):
         for child in node.children:
-            yield from _find_node_interval_items(child, interval)
+            yield from _find_node_interval_subsets_items(child, interval)
+
+
+def _find_node_interval_supersets_items(node: Node,
+                                        interval: Interval) -> Iterator[Item]:
+    if _interval.is_subset_of(interval, node.interval):
+        if node.is_leaf:
+            yield node.item
+        else:
+            for child in node.children:
+                yield from _find_node_interval_supersets_items(child, interval)
